"""Test Numpy array subclasses."""

import numpy as np
import pytest

from segy.arrays import HeaderArray
from segy.arrays import SegyArray
from segy.exceptions import InvalidFieldError
from segy.exceptions import NonSpecFieldError


def test_segy_array_copy() -> None:
    """Test copying a segy array with exact underlying buffer."""
    buffer_expected = np.asarray([0, 1, 2, 3, 4], dtype="uint16").tobytes()

    dtype = np.dtype(
        {
            "names": ["f1", "f2"],
            "offsets": [0, 4],
            "formats": ["uint16", "uint16"],
            "itemsize": 10,
        }
    )
    segy_array = SegyArray(np.frombuffer(buffer_expected, dtype=dtype))
    segy_array_copy = segy_array.copy()

    assert segy_array_copy.tobytes() == buffer_expected
    assert np.may_share_memory(segy_array_copy, segy_array) is False


@pytest.mark.parametrize(
    ("keys", "aliases", "values"),
    [
        (["samples_per_trace", "sample_interval"], ["ns", "interval"], (5, 2)),
        (["samples_per_trace", "sample_interval"], ["Samples", "dt"], (1, 6)),
    ],
)
class TestHeaderAlias:
    """Test header alias get/set item methods and exceptions."""

    def test_getitem_alias(
        self, keys: list[str], aliases: list[str], values: list[int]
    ) -> None:
        """Test getting values by SU and segyio aliases."""
        array = np.asarray(values, dtype="int16")
        buffer = bytearray(array.tobytes())
        dtype = np.dtype({"names": keys, "formats": ["int16", "int16"]})
        struct = np.frombuffer(buffer, dtype=dtype)
        header_array = HeaderArray(struct)

        # str branch tests
        assert header_array[aliases[0]].item() == values[0]
        assert header_array[aliases[1]].item() == values[1]
        with pytest.raises(InvalidFieldError, match="Invalid key"):
            _ = header_array["foo"]
        with pytest.raises(NonSpecFieldError, match="spec does not define this field"):
            _ = header_array["SEGYRevision"]

        # list[str branch] tests
        assert header_array[aliases].item() == values
        with pytest.raises(NonSpecFieldError, match="spec does not define this field"):
            _ = header_array[["SEGYRevision", "ExtSamples"]]

        # test shortcuts (i.e. not using alias)
        assert header_array[keys[0]].item() == values[0]
        assert header_array[keys].item() == values

    def test_setitem_alias(
        self, keys: list[str], aliases: list[str], values: list[int]
    ) -> None:
        """Test setting values by SU and segyio aliases."""
        array = np.zeros(len(values), dtype="int16")
        buffer = bytearray(array.tobytes())
        dtype = np.dtype({"names": keys, "formats": ["int16", "int16"]})
        struct = np.frombuffer(buffer, dtype=dtype)
        header_array = HeaderArray(struct)

        # str branch tests
        header_array[aliases[0]] = values[0]
        header_array[aliases[1]] = values[1]
        assert header_array[aliases].item() == values

        # list[str branch] tests
        values_x2 = tuple(val * 2 for val in values)
        header_array[aliases] = values_x2
        assert header_array[aliases].item() == values_x2

        # test shortcuts (i.e. not using alias)
        values_x3 = tuple(val * 3 for val in values)
        header_array[keys] = values_x3
        assert header_array[aliases].item() == values_x3



"""Test the usage of SegyFile class."""

from __future__ import annotations

import struct
from dataclasses import dataclass
from typing import TYPE_CHECKING

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal
from numpy.testing import assert_array_equal

from segy import SegyFactory
from segy import SegyFile
from segy.config import BinaryHeaderSettings
from segy.config import SegySettings
from segy.exceptions import EndiannessInferenceError
from segy.exceptions import SegyFileSpecMismatchError
from segy.schema import Endianness
from segy.schema import ScalarType
from segy.schema import SegyStandard
from segy.standards import get_segy_standard
from segy.standards.codes import DataSampleFormatCode

if TYPE_CHECKING:
    from typing import Any

    from fsspec import AbstractFileSystem
    from fsspec.implementations.memory import MemoryFileSystem
    from numpy.typing import NDArray

SAMPLE_INTERVAL = 2000
SAMPLES_PER_TRACE = 21
NUM_TRACES = 15

EXPECTED_SAMPLE_LABELS = range(0, SAMPLES_PER_TRACE * SAMPLE_INTERVAL, SAMPLE_INTERVAL)


@dataclass
class SegyFileTestConfig:
    """Configuration container for testing SegyFile class."""

    uri: str
    segy_standard: SegyStandard
    endianness: Endianness
    sample_format: ScalarType
    expected_headers: NDArray[Any]
    expected_samples: NDArray[Any]


def generate_test_trace_data(
    factory: SegyFactory,
    num_traces: int,
) -> tuple[NDArray[np.void], NDArray[Any]]:
    """Generate random header and sample data for testing."""
    rng = np.random.default_rng()
    header_spec = factory.spec.trace.header
    data_spec = factory.spec.trace.data

    header_dtype = header_spec.dtype.newbyteorder("=")
    header_arr = np.empty(num_traces, dtype=header_dtype)
    for field in header_spec.fields:
        random_field_data = rng.uniform(-128, 127, size=num_traces)
        header_arr[field.name] = random_field_data.astype(field.format)

    # Cast to float32 if IBM.
    if data_spec.format == ScalarType.IBM32:
        sample_dtype = np.dtype("float32")
    else:
        sample_dtype = np.dtype(data_spec.format)
    sample_shape = (num_traces, SAMPLES_PER_TRACE)
    sample_arr = np.empty(shape=sample_shape, dtype=sample_dtype)
    random_sample_data = rng.normal(size=sample_shape)
    sample_arr[:] = random_sample_data.astype("float32")

    return header_arr, sample_arr


def generate_test_segy(
    filesystem: AbstractFileSystem,
    segy_standard: SegyStandard = SegyStandard.REV0,
    endianness: Endianness = Endianness.BIG,
    sample_format: ScalarType = ScalarType.IBM32,
    num_extended_text_headers: int = 0,
) -> SegyFileTestConfig:
    """Function for mocking a SEG-Y file with in memory URI."""
    spec = get_segy_standard(segy_standard)
    spec.endianness = endianness
    spec.trace.data.format = sample_format

    factory = SegyFactory(
        spec=spec,
        sample_interval=SAMPLE_INTERVAL,
        samples_per_trace=SAMPLES_PER_TRACE,
    )

    binary_update_dict = {}
    if num_extended_text_headers > 0 and segy_standard > SegyStandard.REV0:
        binary_update_dict["num_extended_text_headers"] = num_extended_text_headers

    text_file_hdr_bytes = factory.create_textual_header()
    bin_file_hdr_bytes = factory.create_binary_header(binary_update_dict)

    headers = factory.create_trace_header_template(NUM_TRACES)
    samples = factory.create_trace_sample_template(NUM_TRACES)

    header_data, sample_data = generate_test_trace_data(factory, NUM_TRACES)
    headers[:] = header_data
    samples[:] = sample_data

    trace_bytes = factory.create_traces(headers, samples)

    uri = f"memory://{segy_standard.name}_{endianness.value}_{sample_format.value}.segy"
    fp = filesystem.open(uri, mode="wb")

    fp.write(text_file_hdr_bytes)
    fp.write(bin_file_hdr_bytes)

    for _ in range(num_extended_text_headers):
        fp.write(factory.create_textual_header())

    fp.write(trace_bytes)

    return SegyFileTestConfig(
        uri, segy_standard, endianness, sample_format, header_data, sample_data
    )


class TestSegyFile:
    """Test the usage of SegyFile class."""

    @pytest.mark.parametrize(
        "standard",
        [SegyStandard.REV0, SegyStandard.REV1, SegyStandard.REV2, SegyStandard.REV21],
    )
    @pytest.mark.parametrize("endianness", [Endianness.BIG, Endianness.LITTLE])
    @pytest.mark.parametrize("sample_format", [ScalarType.IBM32, ScalarType.FLOAT32])
    def test_infer_spec(
        self,
        mock_filesystem: MemoryFileSystem,
        endianness: Endianness,
        standard: SegyStandard,
        sample_format: ScalarType,
    ) -> None:
        """Tests various attributes and methods of a SegyFile with Rev 0 specs."""
        test_config = generate_test_segy(
            filesystem=mock_filesystem,
            segy_standard=standard,
            endianness=endianness,
            sample_format=sample_format,
        )

        segy_file = SegyFile(test_config.uri)

        # Assert spec
        trace_data_spec = segy_file.spec.trace.data
        assert segy_file.spec.segy_standard == test_config.segy_standard
        assert segy_file.spec.endianness == test_config.endianness
        assert trace_data_spec.format == test_config.sample_format

        # Assert attributes
        assert segy_file.num_traces == NUM_TRACES
        assert segy_file.samples_per_trace == SAMPLES_PER_TRACE
        assert segy_file.num_ext_text == 0
        assert_array_equal(segy_file.sample_labels, EXPECTED_SAMPLE_LABELS)

        # Check if JSON-able dict representation is valid
        assert segy_file.spec._repr_json_() == segy_file.spec.model_dump(mode="json")

        # Test the other case where we exactly specify the spec.
        spec_expected = get_segy_standard(standard)
        spec_expected.endianness = endianness
        spec_expected.trace.data.format = sample_format
        segy_file_expected = SegyFile(test_config.uri, spec=spec_expected)

        assert segy_file_expected.spec == segy_file.spec

    def test_text_file_header(
        self, mock_filesystem: MemoryFileSystem, default_text: str
    ) -> None:
        """Test text file header attribute."""
        test_config = generate_test_segy(mock_filesystem)

        segy_file = SegyFile(test_config.uri)

        # Compare first 5 lines because rest is dynamic.
        assert segy_file.text_header[:400] == default_text[:400]

    @pytest.mark.parametrize("standard", [SegyStandard.REV0, SegyStandard.REV1])
    @pytest.mark.parametrize("endianness", [Endianness.BIG, Endianness.LITTLE])
    @pytest.mark.parametrize("sample_format", [ScalarType.IBM32, ScalarType.INT32])
    def test_binary_file_header(
        self,
        mock_filesystem: MemoryFileSystem,
        endianness: Endianness,
        standard: SegyStandard,
        sample_format: ScalarType,
    ) -> None:
        """Test binary file header values."""
        test_config = generate_test_segy(
            filesystem=mock_filesystem,
            segy_standard=standard,
            endianness=endianness,
            sample_format=sample_format,
        )

        segy_file = SegyFile(test_config.uri)
        binary_header = segy_file.binary_header

        expected_sample_format = DataSampleFormatCode[test_config.sample_format.name]
        assert binary_header["sample_interval"] == SAMPLE_INTERVAL
        assert binary_header["orig_sample_interval"] == SAMPLE_INTERVAL
        assert binary_header["samples_per_trace"] == SAMPLES_PER_TRACE
        assert binary_header["orig_samples_per_trace"] == SAMPLES_PER_TRACE
        assert binary_header["data_sample_format"] == expected_sample_format

    @pytest.mark.parametrize("standard", [SegyStandard.REV0, SegyStandard.REV1])
    @pytest.mark.parametrize("endianness", [Endianness.BIG, Endianness.LITTLE])
    def test_trace_header_accessor(
        self,
        mock_filesystem: MemoryFileSystem,
        endianness: Endianness,
        standard: SegyStandard,
    ) -> None:
        """Test trace header accessor and values."""
        test_config = generate_test_segy(
            filesystem=mock_filesystem,
            segy_standard=standard,
            endianness=endianness,
        )

        segy_file = SegyFile(test_config.uri)

        assert_array_equal(segy_file.header[:], test_config.expected_headers)

    @pytest.mark.parametrize("endianness", [Endianness.BIG, Endianness.LITTLE])
    @pytest.mark.parametrize("sample_format", [ScalarType.IBM32, ScalarType.FLOAT32])
    def test_trace_sample_accessor(
        self,
        mock_filesystem: MemoryFileSystem,
        endianness: Endianness,
        sample_format: ScalarType,
    ) -> None:
        """Test trace sample accessor and values."""
        test_config = generate_test_segy(
            filesystem=mock_filesystem,
            endianness=endianness,
            sample_format=sample_format,
        )

        segy_file = SegyFile(test_config.uri)

        assert_array_almost_equal(segy_file.sample[:], test_config.expected_samples)

    @pytest.mark.parametrize("standard", [SegyStandard.REV0, SegyStandard.REV1])
    @pytest.mark.parametrize("endianness", [Endianness.BIG, Endianness.LITTLE])
    @pytest.mark.parametrize("sample_format", [ScalarType.IBM32, ScalarType.UINT8])
    def test_trace_accessor(
        self,
        mock_filesystem: MemoryFileSystem,
        endianness: Endianness,
        standard: SegyStandard,
        sample_format: ScalarType,
    ) -> None:
        """Test trace accessor and values."""
        test_config = generate_test_segy(
            filesystem=mock_filesystem,
            segy_standard=standard,
            endianness=endianness,
            sample_format=sample_format,
        )

        segy_file = SegyFile(test_config.uri)
        traces = segy_file.trace[:]

        assert_array_equal(traces.header, test_config.expected_headers)
        assert_array_almost_equal(traces.sample, test_config.expected_samples)

        # Test random access
        index = [0, 2, 4]
        traces = segy_file.trace[index]
        assert_array_equal(traces.header, test_config.expected_headers[index])
        assert_array_almost_equal(traces.sample, test_config.expected_samples[index])

        # Test reverse order random access
        index = [5, 3, 0]
        traces = segy_file.trace[index]
        assert_array_equal(traces.header, test_config.expected_headers[index])
        assert_array_almost_equal(traces.sample, test_config.expected_samples[index])

        # Test random access with duplicates
        index = [5, 3, 0, 5, 5, 3]
        traces = segy_file.trace[index]
        assert_array_equal(traces.header, test_config.expected_headers[index])
        assert_array_almost_equal(traces.sample, test_config.expected_samples[index])


class TestSegyFileExceptions:
    """Test exceptions for SegyFile."""

    def test_endian_code_err_handling(self, mock_filesystem: MemoryFileSystem) -> None:
        """Test bad values in binary header endian code triggering legacy inference."""
        test_config = generate_test_segy(
            filesystem=mock_filesystem,
            segy_standard=SegyStandard.REV1,
        )

        with mock_filesystem.open(test_config.uri, mode="r+b") as fp:
            fp.seek(3296)
            fp.write(struct.pack(">I", 999))

        file = SegyFile(test_config.uri)

        assert file.spec.endianness == Endianness.BIG
        assert file.spec.segy_standard == SegyStandard.REV1

    def test_endian_infer_failure(self, mock_filesystem: MemoryFileSystem) -> None:
        """Test bad values in binary header triggering spec inference failure."""
        test_config = generate_test_segy(filesystem=mock_filesystem)

        with mock_filesystem.open(test_config.uri, mode="r+b") as fp:
            fp.seek(3224)
            fp.write(struct.pack(">H", 420))  # invalid sample format

        error_message = "Endianness inference failed after attempting all methods."
        with pytest.raises(EndiannessInferenceError, match=error_message):
            SegyFile(test_config.uri)


class TestSegyFileSettingsOverride:
    """Test if settings overrides work fine for SegyFile."""

    def test_revision_override(self, mock_filesystem: MemoryFileSystem) -> None:
        """Make rev0 file and open it as rev1 from settings override."""
        test_config = generate_test_segy(
            filesystem=mock_filesystem, segy_standard=SegyStandard.REV0
        )

        settings = SegySettings.model_validate({"binary": {"revision": 1.0}})
        segy_file = SegyFile(test_config.uri, settings=settings)

        assert segy_file.spec.segy_standard == SegyStandard.REV1
        assert segy_file.binary_header["segy_revision_major"] == 0

    def test_revision_endian_override(self, mock_filesystem: MemoryFileSystem) -> None:
        """Make big-rev0 file and open it as little-rev1 from settings override."""
        test_config = generate_test_segy(
            filesystem=mock_filesystem,
            segy_standard=SegyStandard.REV0,
            endianness=Endianness.BIG,
        )
        # Now ensure overriding endian works. This should raise an
        # error because with little endian the sample format will be
        # interpreted incorrectly from binary header.
        settings_dict_both = {"endianness": "little"}
        settings = SegySettings.model_validate(settings_dict_both)

        with pytest.raises(ValueError, match="is not a valid DataSampleFormatCode"):
            SegyFile(test_config.uri, settings=settings)

    @pytest.mark.parametrize("num_extended_text_headers", [0, 2])
    def test_ext_text_header_override(
        self,
        mock_filesystem: MemoryFileSystem,
        num_extended_text_headers: int,
    ) -> None:
        """Test if settings override for extended header count work for SegyFile."""
        # Create file with zero extended text headers
        test_config = generate_test_segy(
            mock_filesystem,
            segy_standard=SegyStandard.REV1,
            num_extended_text_headers=num_extended_text_headers,
        )

        # Mess up and write 42 headers
        with mock_filesystem.open(test_config.uri, mode="r+b") as fp:
            fp.seek(3504)
            fp.write(struct.pack(">H", 42))  # invalid sample format

        # Ensure failure as is
        error_message = "doesn't match parsed spec"
        with pytest.raises(SegyFileSpecMismatchError, match=error_message):
            SegyFile(test_config.uri)

        # Make it work with override
        bin_override = BinaryHeaderSettings(ext_text_header=num_extended_text_headers)
        settings = SegySettings(binary=bin_override)
        segy_file = SegyFile(test_config.uri, settings=settings)

        assert segy_file.num_ext_text == num_extended_text_headers



"""Tests for segy spec components."""

import numpy as np
import pytest

from segy.schema import HeaderField
from segy.schema import ScalarType
from segy.schema import SegySpec
from segy.schema import TextHeaderEncoding
from segy.schema import TextHeaderSpec
from segy.schema import TraceDataSpec
from segy.schema.text_header import ExtendedTextHeaderSpec
from segy.standards import SegyStandard
from segy.standards import get_segy_standard


@pytest.fixture(params=[SegyStandard.REV0, SegyStandard.REV1])
def segy_spec(request: pytest.FixtureRequest) -> SegySpec:
    """Fixture for creating known SegySpec instances for customizing."""
    standard = request.param
    return get_segy_standard(standard)


class TestSegySpecCustomize:
    """Tests for SegySpec customization."""

    def test_custom_textual_file_header(self, segy_spec: SegySpec) -> None:
        """Test customizing text header spec."""
        custom_text_spec = TextHeaderSpec(
            rows=1,
            cols=5,
            encoding=TextHeaderEncoding.EBCDIC,
        )

        custom_spec = segy_spec.customize(text_header_spec=custom_text_spec)

        assert custom_spec.segy_standard is None
        assert custom_spec.text_header == custom_text_spec

    def test_custom_binary_file_headers(self, segy_spec: SegySpec) -> None:
        """Test customizing binary file header spec."""
        custom_fields = [
            HeaderField(name="f1", format=ScalarType.UINT8, byte=17),
            HeaderField(name="f2", format=ScalarType.INT16, byte=33),
            HeaderField(name="f3", format=ScalarType.UINT32, byte=301),
        ]

        custom_spec = segy_spec.customize(binary_header_fields=custom_fields)

        expected_itemsize = segy_spec.binary_header.dtype.itemsize
        assert custom_spec.segy_standard is None
        assert len(custom_spec.binary_header.fields) == len(custom_fields)
        assert custom_spec.binary_header.dtype.names == ("f1", "f2", "f3")
        assert custom_spec.binary_header.dtype.itemsize == expected_itemsize

    def test_custom_trace_headers(self, segy_spec: SegySpec) -> None:
        """Test customizing trace header spec."""
        custom_fields = [
            HeaderField(name="f1", format=ScalarType.UINT8, byte=9),
            HeaderField(name="f2", format=ScalarType.UINT32, byte=151),
        ]

        custom_spec = segy_spec.customize(trace_header_fields=custom_fields)

        expected_itemsize = segy_spec.trace.header.dtype.itemsize
        assert custom_spec.segy_standard is None
        assert len(custom_spec.trace.header.fields) == len(custom_fields)
        assert custom_spec.trace.header.dtype.names == ("f1", "f2")
        assert custom_spec.trace.header.dtype.itemsize == expected_itemsize

    def test_custom_extended_text_header(self, segy_spec: SegySpec) -> None:
        """Test customizing extended text header spec."""
        custom_text_spec = TextHeaderSpec(
            rows=3,
            cols=7,
            encoding=TextHeaderEncoding.ASCII,
        )

        custom_ext_text_spec = ExtendedTextHeaderSpec(spec=custom_text_spec, count=1)
        custom_spec = segy_spec.customize(ext_text_spec=custom_ext_text_spec)

        assert custom_spec.segy_standard is None
        assert custom_spec.ext_text_header == custom_ext_text_spec

    def test_custom_trace_samples(self, segy_spec: SegySpec) -> None:
        """Test customizing trace data spec."""
        custom_samples = TraceDataSpec(format=ScalarType.UINT16, samples=3)

        custom_spec = segy_spec.customize(trace_data_spec=custom_samples)

        expected_subdtype = (np.dtype("uint16"), (3,))
        assert custom_spec.segy_standard is None
        assert custom_spec.trace.dtype.itemsize == 246  # noqa: PLR2004
        assert custom_spec.trace.data.dtype.itemsize == 6  # noqa: PLR2004
        assert custom_spec.trace.data.dtype.subdtype == expected_subdtype



"""Tests for components that define fields and data types of a SEGY Schema."""

from __future__ import annotations

import pytest

from segy.schema import Endianness
from segy.schema import HeaderField
from segy.schema import HeaderSpec
from segy.schema import ScalarType


class TestHeaderSpec:
    """Tests for header field spec and header spec."""

    @pytest.mark.parametrize("endianness", [Endianness.BIG, Endianness.LITTLE])
    @pytest.mark.parametrize(
        ("names", "bytes_", "formats", "itemsize"),
        [
            (["f1"], [1], [ScalarType.INT32], None),
            (["f1", "f2"], [1, 9], [ScalarType.UINT8, ScalarType.INT32], None),
            (["f1", "f2"], [1, 13], [ScalarType.UINT8, ScalarType.INT16], 20),
        ],
    )
    def test_header_spec(  # noqa: PLR0913
        self,
        endianness: Endianness,
        names: list[str],
        bytes_: list[int],
        formats: list[ScalarType],
        itemsize: int,
    ) -> None:
        """Test HeaderSpec and its methods."""
        fields = [
            HeaderField(name=name, byte=byte, format=format_)
            for name, byte, format_ in zip(names, bytes_, formats, strict=True)
        ]

        header_spec = HeaderSpec(
            fields=fields, item_size=itemsize, endianness=endianness
        )

        assert header_spec.dtype.names == tuple(names)
        if endianness == Endianness.LITTLE:
            assert header_spec.dtype.isnative
        if endianness == Endianness.BIG:
            assert not header_spec.dtype.isnative
        if itemsize is not None:
            assert header_spec.dtype.itemsize == itemsize

    def test_add_field(self) -> None:
        """Test adding fields to header spec with and without overwrite."""
        header_spec = HeaderSpec(fields=[])

        field1 = HeaderField(name="f1", byte=3, format=ScalarType.INT16)
        field2 = HeaderField(name="f2", byte=7, format=ScalarType.INT16)

        header_spec.add_field(field1)
        header_spec.add_field(field2)

        actual_fields = dict(header_spec.dtype.fields)  # type: ignore
        assert header_spec.dtype.itemsize == field2.offset + field2.dtype.itemsize
        assert header_spec.dtype.names == ("f1", "f2")
        assert actual_fields["f1"] == (field1.dtype, field1.offset)
        assert actual_fields["f2"] == (field2.dtype, field2.offset)

        # Test with overwrite.
        field2 = HeaderField(name="f2", byte=11, format=ScalarType.INT32)
        header_spec.add_field(field2, overwrite=True)

        actual_fields = dict(header_spec.dtype.fields)  # type: ignore
        assert header_spec.dtype.itemsize == field2.offset + field2.dtype.itemsize
        assert actual_fields["f1"] == (field1.dtype, field1.offset)
        assert actual_fields["f2"] == (field2.dtype, field2.offset)

    def test_remove_field(self) -> None:
        """Test field removal from header spec."""
        header_spec = HeaderSpec(
            fields=[
                HeaderField(name="f1", byte=11, format=ScalarType.INT8),
                HeaderField(name="f2", byte=21, format=ScalarType.INT16),
            ]
        )

        header_spec.remove_field("f2")

        remaining_field = header_spec.fields[0]
        expected_itemsize = remaining_field.offset + remaining_field.dtype.itemsize
        assert header_spec.itemsize == expected_itemsize
        assert header_spec.dtype.names == ("f1",)

    def test_header_spec_deserialize(self) -> None:
        """Test for validating recreating a header spec from a JSON string."""
        struct_json = """
        {
          "fields": [
            {"description": "field1", "format": "int32", "name": "f1", "byte": 1},
            {"description": "field2", "format": "ibm32", "name": "f2", "byte": 5}
          ],
          "itemSize": 12,
          "offset": 200,
          "endianness": "big"
        }
        """
        actual_spec = HeaderSpec.model_validate_json(struct_json)
        assert len(actual_spec.fields) == 2  # noqa: PLR2004
        assert actual_spec.dtype.names == ("f1", "f2")
        assert actual_spec.dtype.itemsize == 12  # noqa: PLR2004
        assert not actual_spec.dtype.isnative


class TestHeaderSpecExceptions:
    """Test exceptions in header spec."""

    def test_validation_duplicate_field_exception(self) -> None:
        """Test expected failure when multiple keys are provided more than once."""
        fields_with_duplicate = [
            HeaderField(name="f1", byte=3, format=ScalarType.INT16),
            HeaderField(name="f1", byte=9, format=ScalarType.INT16),
        ]
        with pytest.raises(ValueError, match="Duplicate header fields detected"):
            HeaderSpec(fields=fields_with_duplicate)

    def test_add_field_exists_exception(self) -> None:
        """Test adding fields that already exists without overwrite flag."""
        header_spec = HeaderSpec(
            fields=[
                HeaderField(name="f1", byte=3, format=ScalarType.INT16),
                HeaderField(name="f2", byte=11, format=ScalarType.INT16),
            ]
        )

        field = HeaderField(name="f1", byte=7, format=ScalarType.UINT8)
        with pytest.raises(KeyError, match="f1 already exists."):
            header_spec.add_field(field)

    def test_remove_field_missing_exception(self) -> None:
        """Test field removal from structured data type."""
        header_spec = HeaderSpec(
            fields=[
                HeaderField(name="f1", byte=11, format=ScalarType.INT8),
                HeaderField(name="f2", byte=21, format=ScalarType.INT16),
            ]
        )

        with pytest.raises(KeyError, match="f3 does not exist."):
            header_spec.remove_field("f3")

    @pytest.mark.parametrize(
        ("names", "bytes_", "formats", "itemsize"),
        [
            (["f1"], [1], [ScalarType.UINT64], 4),
            (["f1", "f2"], [1, 24], [ScalarType.UINT8, ScalarType.INT16], 16),
        ],
    )
    def test_struct_size_overflow(
        self,
        names: list[str],
        bytes_: list[int],
        formats: list[ScalarType],
        itemsize: int,
    ) -> None:
        """Test validation where the max field size exceeds the item size."""
        fields = [
            HeaderField(name=name, byte=byte, format=format_)
            for name, byte, format_ in zip(names, bytes_, formats, strict=True)
        ]
        with pytest.raises(ValueError, match="Offsets exceed allowed header size."):
            HeaderSpec(fields=fields, item_size=itemsize)



"""Tests for IBM and IEEE floating point conversions.

Some references for test values
https://en.wikipedia.org/wiki/IBM_hexadecimal_floating-point
https://www.crewes.org/Documents/ResearchReports/2017/CRR201725.pdf
"""

import numpy as np
import pytest

from segy.ibm import ibm2ieee
from segy.ibm import ibm2ieee_single
from segy.ibm import ieee2ibm
from segy.ibm import ieee2ibm_single


@pytest.mark.parametrize(
    ("ieee", "ibm"),
    [
        (0.0, 0x00000000),
        (-0.0, 0x00000000),
        (0.1, 0x40199999),
        (-1, 0xC1100000),
        (3.141593, 0x413243F7),
        (-0.15625, 0xC0280000),
        (118.625, 0x4276A000),
        (-8521603, 0xC6820783),
        (3.4028235e38, 0x60FFFFFF),
        (-3.4028235e38, 0xE0FFFFFF),
        ([-0.0, 0.1], [0x00000000, 0x40199999]),
        ([0.0, 0.1, 3.141593], [0x00000000, 0x40199999, 0x413243F7]),
        ([[0.0], [0.1], [3.141593]], [[0x00000000], [0x40199999], [0x413243F7]]),
    ],
)
class TestIbmIeee:
    """Test conversions between values as Ibm floats to Ieee floats."""

    def test_ieee_to_ibm(self, ieee: float, ibm: int) -> None:
        """Test converting values from IEEE to IBM."""
        ieee_float32 = np.float32(ieee)
        actual_ibm = ieee2ibm(ieee_float32)
        expected_ibm = np.uint32(ibm)
        np.testing.assert_array_equal(actual_ibm, expected_ibm)

    def test_ibm_to_ieee(self, ieee: float, ibm: int) -> None:
        """Test converting values from IBM to IEEE."""
        expected_ieee = np.float32(ieee)
        actual_ibm = np.uint32(ibm)
        actual_ieee = ibm2ieee(actual_ibm)
        np.testing.assert_array_almost_equal(actual_ieee, expected_ieee)


@pytest.mark.parametrize(
    ("ieee", "ibm"),
    [
        (0.0, 0x00000000),
        (-0.0, 0x00000000),
        (0.1, 0x40199999),
        (-1, 0xC1100000),
        (3.141593, 0x413243F7),
        (-0.15625, 0xC0280000),
        (118.625, 0x4276A000),
        (-8521603, 0xC6820783),
        (3.4028235e38, 0x60FFFFFF),
        (-3.4028235e38, 0xE0FFFFFF),
    ],
)
class TestSingleIbmIeee:
    """Test conversions for single values from Ibm floats to Ieee floats."""

    def test_single_ieee_to_ibm(self, ieee: float, ibm: int) -> None:
        """Test converting single values from IEEE to IBM."""
        ieee_float32 = np.float32(ieee)
        actual_ibm = ieee2ibm_single(ieee_float32)
        expected_ibm = np.uint32(ibm)
        np.testing.assert_array_almost_equal(actual_ibm, expected_ibm)

    def test_single_ibm_to_ieee(self, ieee: float, ibm: int) -> None:
        """Test converting values from IBM to IEEE."""
        expected_ieee = np.float32(ieee)
        actual_ibm = np.uint32(ibm)
        acutal_ieee = ibm2ieee_single(actual_ibm)
        np.testing.assert_array_almost_equal(acutal_ieee, expected_ieee)


@pytest.mark.parametrize("shape", [(1,), (10,), (20, 20), (150, 150)])
def test_ieee_to_ibm_roundtrip(shape: tuple[int, ...]) -> None:
    """Convert values from IEEE to IBM and back to IEEE."""
    rng = np.random.default_rng()
    expected_ieee = rng.normal(size=shape).astype("float32")
    actual_ibm = ieee2ibm(expected_ieee)
    actual_ieee = ibm2ieee(actual_ibm)

    np.testing.assert_array_almost_equal(actual_ieee, expected_ieee)



"""Testing the header accessor and transforms."""

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import cast

import numpy as np
import pytest

from segy import SegyFactory
from segy.schema import SegyStandard
from segy.schema.base import Endianness
from segy.standards import get_segy_standard
from segy.transforms import TransformFactory
from segy.transforms import TransformPipeline

if TYPE_CHECKING:
    from typing import Any

    from numpy.typing import NDArray


@pytest.fixture
def mock_header_little() -> NDArray[Any]:
    """Generate a mock structured array to test transforms with little endian."""
    names = ["field1", "field2", "field3"]
    formats = ["<u4", "<i4", "<f4"]

    dtype = np.dtype({"names": names, "formats": formats})
    arr = np.empty(shape=1, dtype=dtype)
    arr[:] = (42, 1, 3.1415)

    return arr


@pytest.fixture
def mock_header_big(mock_header_little: NDArray[Any]) -> NDArray[Any]:
    """Generate a mock structured array to test transforms with big endian."""
    data = mock_header_little.byteswap()
    swapped_dtype = data.dtype.newbyteorder()
    return data.view(swapped_dtype)


@pytest.fixture
def mock_header_ibm() -> NDArray[Any]:
    """Generate a mock structured array to test IBM float field transform."""
    names = ["u4_field", "ibm_field", "u2_field", "ibm_field2"]
    formats = ["<u4", "<u4", "<u2", "<u4"]

    dtype = np.dtype({"names": names, "formats": formats})
    arr = np.empty(shape=1, dtype=dtype)
    arr[:] = (256, 0x4276A000, 8, 0x413243F7)

    return arr


@pytest.fixture
def mock_data_little() -> NDArray[Any]:
    """Generate a mock little endian structured array to test transforms."""
    return np.asarray([[3.14, 42, 0], [-2, 3, -4]]).astype(np.float32)


@pytest.fixture
def mock_data_big(mock_data_little: NDArray[Any]) -> NDArray[Any]:
    """Generate a mock big endian structured array to test transforms."""
    data = mock_data_little.byteswap()
    swapped_dtype = data.dtype.newbyteorder()
    return data.view(swapped_dtype)


class TestByteSwap:
    """Test byte swap transformations."""

    @pytest.mark.parametrize(
        ("input_endian", "expected_order"),
        [
            ("little", Endianness.BIG),
            ("big", Endianness.LITTLE),
            ("big", Endianness.BIG),  # test_noop
        ],
    )
    def test_byteswap(
        self,
        request: pytest.FixtureRequest,
        input_endian: str,
        expected_order: Endianness,
    ) -> None:
        """Test byte swap transform."""
        mock_data = request.getfixturevalue(f"mock_data_{input_endian}")
        expected_dtype = mock_data.dtype.newbyteorder(expected_order.symbol)

        # Transform but make copy before because transform is in place.
        transform = TransformFactory.create("byte_swap", expected_order)
        swapped_data = transform.apply(mock_data.copy())

        np.testing.assert_allclose(swapped_data, mock_data)
        assert swapped_data.dtype == expected_dtype


class TestIbmFloat:
    """Test IBM Float transformations."""

    def test_ibm_float(self) -> None:
        """Test array scaling."""
        ieee_value = np.asarray(3.141593)
        expected_ibm_uint32 = 0x413243F7

        transform = TransformFactory.create("ibm_float", "to_ibm")
        ibm_uint32 = transform.apply(ieee_value)

        assert ibm_uint32 == expected_ibm_uint32

        ibm_value = np.asarray([0x4276A000], dtype="uint32")
        expected_ieee = np.asarray([118.625])

        transform = TransformFactory.create("ibm_float", "to_ieee")
        ibm_value = transform.apply(ibm_value)

        np.testing.assert_array_equal(ibm_value, expected_ieee)

    def test_ibm_float_field(self, mock_header_ibm: NDArray[Any]) -> None:
        """Test array scaling."""
        expected_dtype = [
            ("", "uint32"),
            ("", "float32"),
            ("", "uint16"),
            ("", "float32"),
        ]
        expected = np.asarray((256, 118.625, 8, 3.141593), dtype=expected_dtype)

        transform = TransformFactory.create(
            "ibm_float", "to_ieee", keys=["ibm_field", "ibm_field2"]
        )
        transformed_header = transform.apply(mock_header_ibm)

        assert transformed_header[0].item() == expected.item()


class TestRevisionTransform:
    """Test SEG-Y revision transforms. This tests both transform and factory."""

    @pytest.mark.parametrize("endian", ["little", "big"])
    @pytest.mark.parametrize(("major", "minor"), [(0, 0), (1, 0), (2, 0)])
    def test_rev_parse(self, endian: str, major: int, minor: int) -> None:
        """Test array scaling."""
        # Set parameters
        rev_float = float(f"{major}.{minor}")
        revision = SegyStandard(rev_float)
        endianness = Endianness(endian)

        # Create binary header with factory
        spec = get_segy_standard(revision)
        spec.endianness = endianness
        segy_factory = SegyFactory(spec)
        bin_header_bytes = bytearray(segy_factory.create_binary_header())
        bin_header = np.frombuffer(bin_header_bytes, dtype=spec.binary_header.dtype)

        # Set up and apply transform, expects native endian
        if endian == "big":
            bin_header = bin_header.byteswap()
            bin_header = bin_header.view(spec.binary_header.dtype.newbyteorder())

        transform = TransformFactory.create("segy_revision")
        transformed_bin_header = transform.apply(bin_header)

        header_fields = cast("tuple[str]", transformed_bin_header.dtype.names)

        assert transformed_bin_header["segy_revision_major"].squeeze() == major
        assert transformed_bin_header["segy_revision_minor"].squeeze() == minor
        assert "segy_revision" not in header_fields


class TestTransformPipeline:
    """Tests for transform pipeline and transform integration."""

    def test_transform_pipeline(self, mock_data_little: NDArray[Any]) -> None:
        """Test transformation pipeline and transform integration."""
        expected_data = mock_data_little.copy()

        pipeline = TransformPipeline()
        pipeline.add_transform(TransformFactory.create("byte_swap", Endianness.BIG))
        pipeline.add_transform(TransformFactory.create("byte_swap", Endianness.LITTLE))

        transformed_data = pipeline.apply(mock_data_little)
        np.testing.assert_allclose(transformed_data, expected_data)
        assert transformed_data.dtype == expected_data.dtype

    def test_transform_pipeline_field(self, mock_header_ibm: NDArray[Any]) -> None:
        """Test transformation pipeline and transform integration for struct fields."""
        expected_dtype = [
            ("u4_field", "uint32"),
            ("ibm_field", "float32"),
            ("u2_field", "uint16"),
            ("ibm_field2", "float32"),
        ]
        expected_data = np.asarray((256, 118.625, 8, 3.141593), dtype=expected_dtype)

        keys = ["ibm_field", "ibm_field2"]
        pipeline = TransformPipeline()
        pipeline.add_transform(TransformFactory.create("byte_swap", Endianness.BIG))
        pipeline.add_transform(TransformFactory.create("byte_swap", Endianness.LITTLE))
        pipeline.add_transform(TransformFactory.create("ibm_float", "to_ieee", keys))

        transformed_data = pipeline.apply(mock_header_ibm.copy())

        assert transformed_data == expected_data
        assert transformed_data.dtype == expected_dtype

    def test_transform_exceptions(
        self,
        mock_data_little: NDArray[Any],
        mock_header_big: NDArray[Any],
    ) -> None:
        """Test transformation field errors."""
        scale_transform = TransformFactory.create("ibm_float", "to_ibm", ["field1"])
        with pytest.raises(ValueError, match="doesn't contain any fields"):
            scale_transform.apply(mock_data_little)

        scale_transform = TransformFactory.create("ibm_float", "to_ibm", ["unknown"])
        with pytest.raises(ValueError, match="{'unknown'} not in the array"):
            scale_transform.apply(mock_header_big)

    def test_transform_factory_exception(self) -> None:
        """Test unknown transformation type."""
        factory = TransformFactory()

        with pytest.raises(KeyError, match="Unsupported transformation"):
            factory.create("non_existent_transform")



"""Tests for functions in segy.schema.registry."""

import pytest

from segy.schema import SegySpec
from segy.standards import SegyStandard
from segy.standards import get_segy_standard
from segy.standards.spec import REV0
from segy.standards.spec import REV1


@pytest.mark.parametrize(
    ("standard_enum", "standard_spec"),
    [(SegyStandard.REV0, REV0), (SegyStandard.REV1, REV1)],
)
def test_get_standard(standard_enum: SegyStandard, standard_spec: SegySpec) -> None:
    """Test retrieving SegyStandard from registry."""
    spec_copy = get_segy_standard(standard_enum)
    assert spec_copy == standard_spec
    assert id(spec_copy) != id(standard_spec)


def test_get_nonexistent_spec_error() -> None:
    """Test missing / non-existent SegyStandard from registry."""
    with pytest.raises(NotImplementedError):
        get_segy_standard("non_existent")



"""Tests for the SegyFactory used in creation."""

from __future__ import annotations

import random
import string
from dataclasses import dataclass

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from segy.factory import SegyFactory
from segy.ibm import ieee2ibm
from segy.schema import Endianness
from segy.schema import HeaderField
from segy.schema import ScalarType
from segy.schema import SegyStandard
from segy.schema import TextHeaderEncoding
from segy.standards.codes import DataSampleFormatCode
from segy.standards.minimal import minimal_segy


@dataclass
class SegyFactoryTestConfig:
    """Dataclass to configure common test patterns."""

    segy_standard: SegyStandard | None
    endianness: Endianness
    sample_interval: int
    samples_per_trace: int


SEGY_FACTORY_TEST_CONFIGS = [
    SegyFactoryTestConfig(None, Endianness.BIG, 2000, 51),
    SegyFactoryTestConfig(SegyStandard.REV0, Endianness.BIG, 2000, 51),
    SegyFactoryTestConfig(SegyStandard.REV1, Endianness.LITTLE, 3000, 1),
    SegyFactoryTestConfig(SegyStandard.REV0, Endianness.BIG, 5000, 10),
]


@pytest.fixture(params=SEGY_FACTORY_TEST_CONFIGS)
def mock_segy_factory(request: pytest.FixtureRequest) -> SegyFactory:
    """Generates the test cases for SegyFactory.

    We start with a minimal SEG-Y spec and modify its properties to generate various
    files parametrically. `SEGY_FACTORY_TEST_CONFIGS` defines the base configuration.
    """
    test_config = request.param
    spec = minimal_segy

    # Set file wide attributes
    spec.endianness = test_config.endianness
    spec.segy_standard = test_config.segy_standard

    # Shrink trace headers to 16-bytes and add a few fields
    spec.trace.header.item_size = 32
    spec.trace.header.fields = [
        HeaderField(name="field1", format=ScalarType.INT8, byte=3),
        HeaderField(name="field2", format=ScalarType.INT32, byte=5),
        HeaderField(name="field3", format=ScalarType.UINT8, byte=11),
    ]

    return SegyFactory(
        spec,
        sample_interval=test_config.sample_interval,
        samples_per_trace=test_config.samples_per_trace,
    )


@pytest.mark.parametrize(
    "encoding", [TextHeaderEncoding.EBCDIC, TextHeaderEncoding.ASCII]
)
class TestTextualFileHeader:
    """Tests related to the text header writing."""

    def test_text_header_encoding(
        self, encoding: TextHeaderEncoding, default_text: str
    ) -> None:
        """Tests the text header encoding/decoding works correctly."""
        segy_spec = minimal_segy
        segy_spec.text_header.encoding = encoding
        factory = SegyFactory(segy_spec)

        text_bytes = factory.create_textual_header()
        text_actual = segy_spec.text_header.decode(text_bytes)

        # Compare first 5 lines because rest is dynamic.
        assert text_actual[:400] == default_text[:400]

    def test_text_header_custom(self, encoding: TextHeaderEncoding) -> None:
        """Test if custom header is encoded/decoded correctly."""
        segy_spec = minimal_segy
        segy_spec.text_header.encoding = encoding
        factory = SegyFactory(segy_spec)

        chars = string.ascii_uppercase + string.ascii_lowercase + string.ascii_letters
        text_rows = []
        for _ in range(40):
            row_string = "".join([random.choice(chars) for _ in range(80)])  # noqa: S311
            text_rows.append(row_string)
        text_expected = "\n".join(text_rows)

        text_bytes = factory.create_textual_header(text_expected)
        text_actual = segy_spec.text_header.decode(text_bytes)

        assert text_actual == text_expected


class TestBinaryFileHeader:
    """Tests related to the text header writing."""

    @pytest.mark.parametrize(
        "sample_format", [ScalarType.FLOAT32, ScalarType.IBM32, ScalarType.INT16]
    )
    def test_binary_header_default(
        self, mock_segy_factory: SegyFactory, sample_format: ScalarType
    ) -> None:
        """Ensure the binary header is properly encoded and serialized."""
        mock_segy_factory.spec.trace.data.format = sample_format

        binary_bytes = mock_segy_factory.create_binary_header()

        bin_spec = mock_segy_factory.spec.binary_header
        binary_actual = np.frombuffer(binary_bytes, dtype=bin_spec.dtype)
        binary_expected = (
            mock_segy_factory.sample_interval,
            mock_segy_factory.sample_interval,
            mock_segy_factory.samples_per_trace,
            mock_segy_factory.samples_per_trace,
            DataSampleFormatCode[mock_segy_factory.sample_format.name],
            mock_segy_factory.segy_revision.value * 256,
            0,  # fixed length trace flag
            0,  # extended text headers
        )
        assert binary_actual.item() == binary_expected

    def test_binary_header_custom(self, mock_segy_factory: SegyFactory) -> None:
        """Test if binary header encoding field update."""
        mock_segy_factory.spec.trace.data.format = ScalarType.IBM32

        update = {
            "fixed_length_trace_flag": 1,
            "num_extended_text_headers": 2,
        }
        binary_bytes = mock_segy_factory.create_binary_header(update=update)

        bin_spec = mock_segy_factory.spec.binary_header
        binary_actual = np.frombuffer(binary_bytes, dtype=bin_spec.dtype)
        binary_expected = (
            mock_segy_factory.sample_interval,
            mock_segy_factory.sample_interval,
            mock_segy_factory.samples_per_trace,
            mock_segy_factory.samples_per_trace,
            DataSampleFormatCode[mock_segy_factory.sample_format.name],
            mock_segy_factory.segy_revision.value * 256,
            1,  # fixed length trace flag
            2,  # extended text headers
        )
        assert binary_actual.item() == binary_expected


@pytest.mark.parametrize("num_traces", [1, 42])
class TestSegyFactoryTraces:
    """Ensure the trace headers are properly encoded and serialized."""

    def test_trace_header_template(
        self, mock_segy_factory: SegyFactory, num_traces: int
    ) -> None:
        """Test if the trace header template is correct."""
        headers = mock_segy_factory.create_trace_header_template(num_traces)

        header_spec = mock_segy_factory.spec.trace.header
        assert headers.size == num_traces
        assert headers.dtype == header_spec.dtype.newbyteorder("<")

    def test_trace_header_template_with_sample_info(
        self, mock_segy_factory: SegyFactory, num_traces: int
    ) -> None:
        """Test if the trace header template is correct with sample info."""
        mock_segy_factory.spec.trace.header.item_size = 26
        # fmt: off
        mock_segy_factory.spec.trace.header.fields += [
            HeaderField(name="sample_interval", format=ScalarType.INT16, byte=13),
            HeaderField(name="samples_per_trace", format=ScalarType.INT16, byte=25),
        ]
        # fmt: on

        headers = mock_segy_factory.create_trace_header_template(num_traces)

        header_spec = mock_segy_factory.spec.trace.header
        expected_sample_info = mock_segy_factory.sample_interval
        expected_samples_per_trace = mock_segy_factory.samples_per_trace
        assert headers.size == num_traces
        assert headers.dtype == header_spec.dtype.newbyteorder("<")
        assert_array_equal(headers["sample_interval"], expected_sample_info)
        assert_array_equal(headers["samples_per_trace"], expected_samples_per_trace)

    @pytest.mark.parametrize(
        "sample_format", [ScalarType.FLOAT32, ScalarType.IBM32, ScalarType.INT16]
    )
    def test_trace_sample_template(
        self, mock_segy_factory: SegyFactory, num_traces: int, sample_format: ScalarType
    ) -> None:
        """Test if the trace sample template is correct."""
        mock_segy_factory.spec.trace.data.format = sample_format

        samples = mock_segy_factory.create_trace_sample_template(num_traces)

        n_samples = mock_segy_factory.samples_per_trace

        if mock_segy_factory.sample_format == ScalarType.IBM32:
            expected_dtype = np.dtype("float32")
        else:
            expected_dtype = mock_segy_factory.sample_format.dtype

        expected_shape = (num_traces, n_samples)
        assert samples.dtype == expected_dtype
        assert samples.shape == expected_shape

    @pytest.mark.parametrize(
        "sample_format", [ScalarType.FLOAT32, ScalarType.IBM32, ScalarType.INT16]
    )
    def test_trace_serialize(
        self, mock_segy_factory: SegyFactory, num_traces: int, sample_format: ScalarType
    ) -> None:
        """Test if the trace serialization is correct."""
        mock_segy_factory.spec.trace.data.format = sample_format

        # Generate random data
        rng = np.random.default_rng()
        samples_per_trace = mock_segy_factory.samples_per_trace
        shape = (num_traces, samples_per_trace)
        rand_samples = np.float32(255 * rng.random(size=shape))
        rand_fields = {}
        for field in mock_segy_factory.spec.trace.header.fields:
            field_data = rng.integers(-100, 100, dtype="int8", size=num_traces)
            rand_fields[field.name] = field_data

        # Fill in actual and generate bytes
        headers = mock_segy_factory.create_trace_header_template(num_traces)
        samples = mock_segy_factory.create_trace_sample_template(num_traces)
        samples[:] = rand_samples
        for field_name, values in rand_fields.items():
            headers[field_name] = values
        trace_bytes = mock_segy_factory.create_traces(headers, samples)

        # Fill in expected and assertions
        # 1. handle ibm float
        # 2. fill in trace struct
        # 3. handle endianness
        if mock_segy_factory.sample_format == ScalarType.IBM32:
            rand_samples = ieee2ibm(rand_samples)
        trace_dtype_native = mock_segy_factory.spec.trace.dtype.newbyteorder("=")
        expected_traces = np.zeros(shape=num_traces, dtype=trace_dtype_native)
        expected_traces["data"] = rand_samples
        for field_name, values in rand_fields.items():
            expected_traces["header"][field_name] = values
        if mock_segy_factory.spec.endianness == Endianness.BIG:
            expected_traces = expected_traces.byteswap(inplace=True)
            expected_dtype = expected_traces.dtype.newbyteorder(">")
            expected_traces = expected_traces.view(expected_dtype)

        assert trace_bytes == expected_traces.tobytes()


class TestSegyFactoryExceptions:
    """Test the exceptions to SegyFactory creation methods."""

    def test_create_trace_incorrect_ndim(self) -> None:
        """Check if trace dimensions are wrong."""
        spec = minimal_segy
        factory = SegyFactory(spec, sample_interval=2, samples_per_trace=5)

        header_1d = factory.create_trace_header_template(5)
        array_4d = np.empty(shape=(5, 5, 5, 5))

        with pytest.raises(AttributeError, match="Data array must be 2-dimensional"):
            factory.create_traces(header_1d, array_4d)

    def test_create_sample_num_samples_mismatch(self) -> None:
        """Check if trace number of samples are wrong."""
        spec = minimal_segy
        factory = SegyFactory(spec, sample_interval=2, samples_per_trace=5)

        header_1d = factory.create_trace_header_template(size=5)
        array_2d = np.empty(shape=(5, 11))

        with pytest.raises(ValueError, match="Trace length must be"):
            factory.create_traces(header_1d, array_2d)

    def test_create_header_sample_mismatch(self) -> None:
        """Check if headers and traces are different sizes."""
        spec = minimal_segy
        factory = SegyFactory(spec, sample_interval=2, samples_per_trace=11)

        header_1d = factory.create_trace_header_template(size=5)
        array_2d = factory.create_trace_sample_template(size=7)

        with pytest.raises(ValueError, match="same number of rows as data array"):
            factory.create_traces(header_1d, array_2d)



"""Tests for the CLI."""

from __future__ import annotations

import os

import pytest
from typer.testing import CliRunner

from segy.cli.segy import app

runner = CliRunner()


@pytest.fixture
def s3_path() -> str:
    """Fixture for Stratton dataset on S3 (SEG Wiki)."""
    return "s3://open.source.geoscience/open_data/stratton/segy/navmerged/swath_1_geometry.sgy"


class TestDump:
    """Test class for CLI's dump options."""

    @classmethod
    def setup_class(cls: type[TestDump]) -> None:
        """Set environment variable for anon access to S3."""
        os.environ["SEGY__STORAGE_OPTIONS"] = '{"anon": true}'

    def test_info_dump(self, s3_path: str) -> None:
        """Test generic info dump."""
        result = runner.invoke(app, ["dump", "info", s3_path])
        assert result.exit_code == 0
        assert "numTraces" in result.stdout
        assert "fileSize" in result.stdout

    def test_text_dump(self, s3_path: str) -> None:
        """Test text header dump."""
        result = runner.invoke(app, ["dump", "text-header", s3_path])
        assert result.exit_code == 0
        assert "CLIENT: BUREAU OF ECONOMIC GEOLOGY" in result.stdout

    def test_binary_header_dump(self, s3_path: str) -> None:
        """Test binary header dump."""
        result = runner.invoke(app, ["dump", "binary-header", s3_path])
        assert result.exit_code == 0
        assert "sample_interval" in result.stdout
        assert "samples_per_trace" in result.stdout

    def test_trace_header_dump(self, s3_path: str) -> None:
        """Test trace header dump."""
        args = ["dump", "trace-header", s3_path]
        args += ["--index", "100", "--index", "101"]
        args += ["--field", "source_coord_x"]
        args += ["--field", "coordinate_scalar"]

        result = runner.invoke(app, args)
        assert result.exit_code == 0
        assert "source_coord_x" in result.stdout
        assert "coordinate_scalar" in result.stdout
        assert "101" in result.stdout
        assert "70628086" in result.stdout
        assert "-100" in result.stdout

    def test_trace_data_dump(self, s3_path: str) -> None:
        """Test trace data dump."""
        args = ["dump", "trace-data", s3_path]
        args += ["--index", "501", "--index", "1000"]

        result = runner.invoke(app, args)
        assert result.exit_code == 0
        assert "-5.3372304e-08" in result.stdout
        assert "4.2979627e-07" in result.stdout



"""Tests for the Text Headers, Binary Headers, and Trace Headers for different SEGY revisions."""

from __future__ import annotations

import numpy as np
import pytest

from segy.schema import TextHeaderEncoding
from segy.schema import TextHeaderSpec
from segy.schema.text_header import ExtendedTextHeaderSpec


class TestTextHeaderSpec:
    """Tests for the text header spec initialization and encoding/decoding."""

    @pytest.mark.parametrize(
        ("rows", "cols", "encoding", "text"),
        [
            (2, 5, TextHeaderEncoding.ASCII, "hello\nworld"),
            (3, 6, TextHeaderEncoding.EBCDIC, "hello,\n world\nebcdic"),
        ],
    )
    def test_encode_decode(
        self, rows: int, cols: int, encoding: TextHeaderEncoding, text: str
    ) -> None:
        """Test reading ASCII or EBCDIC text headers."""
        text_spec = TextHeaderSpec(rows=rows, cols=cols, encoding=encoding)

        text_bytes = text_spec.encode(text)
        text_roundtrip = text_spec.decode(text_bytes)

        num_char = rows * cols
        assert text_spec.dtype == np.dtype(("uint8", (num_char,)))
        assert text_roundtrip == text


class TestExtTextHeaderSpec:
    """Tests for the text header spec initialization and encoding/decoding."""

    @pytest.mark.parametrize(
        ("rows", "cols", "encoding", "text"),
        [
            (2, 7, TextHeaderEncoding.ASCII, ["hello, \nworld! "]),
            (2, 4, TextHeaderEncoding.EBCDIC, ["foo,\nbar.", "fizz\nbuzz"]),
        ],
    )
    def test_encode_decode(
        self, rows: int, cols: int, encoding: TextHeaderEncoding, text: list[str]
    ) -> None:
        """Test reading ASCII or EBCDIC text headers."""
        count = len(text)

        text_spec = TextHeaderSpec(rows=rows, cols=cols, encoding=encoding)
        ext_text_spec = ExtendedTextHeaderSpec(spec=text_spec, count=count)

        text_bytes = ext_text_spec.encode(text)
        text_roundtrip = ext_text_spec.decode(text_bytes)

        num_char = rows * cols * count
        assert ext_text_spec.dtype == np.dtype(("uint8", (num_char,)))
        assert text_roundtrip == text



"""Tests for indexing.

The indexer classes are tested via the `SegyFile` tests so no
need to test it here.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pytest

from segy.indexing import bounds_check
from segy.indexing import merge_cat_file

if TYPE_CHECKING:
    from fsspec.implementations.memory import MemoryFileSystem


class TestBoundsCheckIndices:
    """Tests for bounds_check function."""

    @pytest.mark.parametrize(
        ("indices", "size"),
        [
            ([1, 2, 3], 5),  # within bounds
            ([4], 5),  # max index (upper edge case)
            ([0], 5),  # min index (lower edge case)
            ([], 5),  # empty index list
        ],
    )
    def test_in_bounds(self, indices: list[int], size: int) -> None:
        """Test the case where indices are in bounds."""
        bounds_check(np.asarray(indices), size, "trace")

    @pytest.mark.parametrize(
        ("indices", "size"),
        [
            ([0, 16], 5),  # out of bounds item (beyond upper limit)
            ([-1], 5),  # negative index (unsupported)
        ],
    )
    def test_out_of_bounds(self, indices: list[int], size: int) -> None:
        """Test the case where indices out of bounds or negative."""
        with pytest.raises(IndexError, match="out of bounds"):
            bounds_check(np.asarray(indices), size, "")


class TestMergeCatFile:
    """Tests for merge_cat_file function."""

    def test_merge_cat_file(self, mock_filesystem: MemoryFileSystem) -> None:
        """Test reading the correct ranges from fake binary data."""
        test_data = b"This is a test file containing test data."
        uri = "/text_merge_cat_file.bin"

        file = mock_filesystem.open(uri, mode="wb")
        file.write(test_data)

        starts = [0, 10, 24]
        ends = [9, 23, 33]

        result = merge_cat_file(mock_filesystem, uri, starts, ends)

        expected_result = b"".join(
            test_data[s:e] for s, e in zip(starts, ends, strict=True)
        )
        assert result == expected_result



"""Tests for trace spec."""

from __future__ import annotations

import pytest

from segy.schema import Endianness
from segy.schema import HeaderField
from segy.schema import HeaderSpec
from segy.schema import ScalarType
from segy.schema import TraceDataSpec
from segy.schema import TraceSpec


@pytest.mark.parametrize("endianness", [Endianness.BIG, Endianness.LITTLE])
@pytest.mark.parametrize(
    ("sample_format", "samples_per_trace"),
    [
        (ScalarType.IBM32, 11),
        (ScalarType.FLOAT32, 21),
        (ScalarType.UINT16, 5),
    ],
)
class TestTraceSpec:
    """Tests for trace spec and its dtype."""

    def test_trace_spec(
        self,
        endianness: Endianness,
        sample_format: ScalarType,
        samples_per_trace: int,
    ) -> None:
        """Testing dtype attribute and contents of TraceSpec."""
        header_spec = HeaderSpec(
            fields=[
                HeaderField(name="h1", format=ScalarType.INT16, byte=1),
                HeaderField(name="h2", format=ScalarType.INT8, byte=17),
            ]
        )
        data_spec = TraceDataSpec(format=sample_format, samples=samples_per_trace)
        trace_spec = TraceSpec(
            header=header_spec,
            data=data_spec,
            endianness=endianness,
        )

        expected_itemsize = header_spec.dtype.itemsize + data_spec.dtype.itemsize
        expected_header_itemsize = header_spec.dtype.itemsize
        expected_sample_subtype = (sample_format.dtype, (samples_per_trace,))
        assert trace_spec.dtype.itemsize == expected_itemsize
        assert trace_spec.header.dtype.names == ("h1", "h2")
        assert trace_spec.header.dtype.itemsize == expected_header_itemsize
        assert trace_spec.data.dtype.subdtype == expected_sample_subtype
        if endianness == Endianness.LITTLE:
            assert trace_spec.dtype.isnative
        else:
            assert not trace_spec.dtype.isnative



"""Tests for accessors for SegyArrays."""

from __future__ import annotations

import pytest

from segy.accessors import TraceAccessor
from segy.schema import Endianness
from segy.schema import HeaderField
from segy.schema import HeaderSpec
from segy.schema import ScalarType
from segy.schema import TraceDataSpec
from segy.schema import TraceSpec
from segy.transforms import Transform
from segy.transforms import TransformFactory
from segy.transforms import TransformPipeline


def compare_transform(transform_a: Transform, transform_b: Transform) -> bool:
    """Helper function for equality comparison of transforms."""
    return all(
        a == b
        for a, b in [
            (type(transform_a), type(transform_b)),
            (transform_a.keys, transform_b.keys),
        ]
    )


def compare_transform_sequence(
    transform_a: list[Transform], transform_b: list[Transform]
) -> bool:
    """Helper function for equality comparison of two sequences of transforms."""
    if len(transform_a) != len(transform_b):
        return False
    return all(
        compare_transform(a, b) for a, b in zip(transform_a, transform_b, strict=True)
    )


@pytest.fixture(
    params=[(ScalarType.IBM32, ScalarType.IBM32), (ScalarType.INT32, ScalarType.IBM32)]
)
def mock_trace_spec(
    request: pytest.FixtureRequest,
) -> tuple[TraceSpec, dict[str, list[Transform]]]:
    """Create mock trace spec for testing and expected values for accessors."""
    header_field_type = request.param[0]
    sample_field_type = request.param[1]
    trace_header_spec = HeaderSpec(
        fields=[HeaderField(name="h1", format=header_field_type, byte=1)],
        item_size=4,
        endianness=Endianness.BIG,
        offset=0,
    )
    trace_data_spec = TraceDataSpec(format=sample_field_type, samples=3)
    trace_spec = TraceSpec(header=trace_header_spec, data=trace_data_spec)
    expected = {}
    base_transform = TransformFactory.create("byte_swap", Endianness.LITTLE)
    expected["header"] = (
        [base_transform]
        if header_field_type != ScalarType.IBM32
        else [base_transform, TransformFactory.create("ibm_float", "to_ieee", ["h1"])]
    )
    expected["data"] = (
        [base_transform]
        if sample_field_type != ScalarType.IBM32
        else [base_transform, TransformFactory.create("ibm_float", "to_ieee")]
    )

    hdr_pipeline = TransformPipeline()
    map(hdr_pipeline.add_transform, expected["header"])
    data_pipeline = TransformPipeline()
    map(data_pipeline.add_transform, expected["data"])
    expected["trace"] = [TransformFactory.create("trace", hdr_pipeline, data_pipeline)]

    return trace_spec, expected


def test_trace_accessor(
    mock_trace_spec: tuple[TraceSpec, dict[str, list[Transform]]],
) -> None:
    """Test for trace accessor decoder transforms."""
    trace_spec = mock_trace_spec[0]
    expected_transforms = mock_trace_spec[1]
    trace_accessor = TraceAccessor(trace_spec)
    assert compare_transform_sequence(
        trace_accessor.header_decode_pipeline.transforms, expected_transforms["header"]
    )
    assert compare_transform_sequence(
        trace_accessor.sample_decode_pipeline.transforms, expected_transforms["data"]
    )
    assert compare_transform_sequence(
        trace_accessor.trace_decode_pipeline.transforms, expected_transforms["trace"]
    )
